import { pipe } from 'fp-ts/function';
import { Mix, Unit, type, $, Vec3, VertexPosition, VertexNormal, Float } from 'shader-composer';
import { PatchedMaterialRoot } from '@material-composer/patch-material';

function _taggedTemplateLiteral(strings, raw) {
  if (!raw) {
    raw = strings.slice(0);
  }
  return Object.freeze(Object.defineProperties(strings, {
    raw: {
      value: Object.freeze(raw)
    }
  }));
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;
  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];
  return arr2;
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

/**
 * Compiles a list of Material Composer modules into a shader graph that
 * can be consumed by Shader Composer's `composeShader` function.
 *
 * @param modules A list of Material Composer modules (see `Module`)
 * @returns A shader root node that can be passed to `compileShader`
 */
var compileModules = function compileModules(modules) {
  /* Transform state with given modules. */
  var state = pipeModules.apply(void 0, [initialModuleState()].concat(_toConsumableArray(modules)));

  /* Construct a shader root unit */
  return PatchedMaterialRoot(state);
};

var _templateObject$1;
var blendFunction = function blendFunction(blend) {
  return function (a, b, f) {
    return f === 0 ? a : blend(a, b, f);
  };
};

/* TODO: implement additional blend modes */

/**
 * Provides a numerb of different blend modes that can be used with a {@link Layer}.
 */
var Blend = {
  normal: blendFunction(function (a, b, f) {
    return f === 1 ? b : f === 0 ? a : Mix(a, b, f);
  }),
  discard: blendFunction(function (a) {
    return a;
  }),
  add: blendFunction(function (a, b, f) {
    return Unit(type(a), $(_templateObject$1 || (_templateObject$1 = _taggedTemplateLiteral(["min(", " + ", ", 1.0) * ", " + ", " * (1.0 - ", ")"])), a, b, f, a, f));
  })
};
/**
 * Layers allow you to group any number of modules (see {@link Module}) and (optionally)
 * apply a blend mode and an opacity to them. Blending will have different effects on different
 * parts of the state; for example, the _blend mode_ will only apply to the color channel, while
 * position, normal and so on will simply be interpolated between according to the configured opacity.
 *
 * @param props - The arguments to the layer. See {@link LayerArgs}.
 */
var Layer = function Layer(_ref) {
  var _ref$modules = _ref.modules,
    modules = _ref$modules === void 0 ? [] : _ref$modules,
    _ref$opacity = _ref.opacity,
    opacity = _ref$opacity === void 0 ? 1 : _ref$opacity,
    _ref$blend = _ref.blend,
    blend = _ref$blend === void 0 ? Blend.normal : _ref$blend;
  return function (state) {
    /* Determine new state */
    var newState = pipeModules.apply(void 0, [state].concat(_toConsumableArray(modules)));

    /* Determine blend function */
    var blendFun = typeof blend === "string" ? Blend[blend] : blendFunction(blend);
    return {
      color: blendFun(state.color, newState.color, opacity),
      position: Blend.normal(state.position, newState.position, opacity),
      alpha: Blend.normal(state.alpha, newState.alpha, opacity),
      normal: Blend.normal(state.normal, newState.normal, opacity),
      roughness: Blend.normal(state.roughness, newState.roughness, opacity),
      metalness: Blend.normal(state.metalness, newState.metalness, opacity)
    };
  };
};

var _templateObject, _templateObject2, _templateObject3, _templateObject4;

/**
 * ModuleState describes the state going into a module (and returned by it.)
 * Modules are encouraged to change the values they're interested in, but can
 * also just pass through others without changing them.
 */

var pipeModules = function pipeModules(initial) {
  for (var _len = arguments.length, modules = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    modules[_key - 1] = arguments[_key];
  }
  return pipe.apply(void 0, [initial].concat(_toConsumableArray(modules)));
};
var initialModuleState = function initialModuleState() {
  return {
    position: Vec3(VertexPosition, {
      name: "Final Position"
    }),
    normal: Vec3(VertexNormal, {
      name: "Final Normal "
    }),
    color: Vec3($(_templateObject || (_templateObject = _taggedTemplateLiteral(["patched_Color"]))), {
      name: "Final Color"
    }),
    alpha: Float($(_templateObject2 || (_templateObject2 = _taggedTemplateLiteral(["patched_Alpha"]))), {
      name: "Final Alpha"
    }),
    roughness: Float($(_templateObject3 || (_templateObject3 = _taggedTemplateLiteral(["patched_Roughness"]))), {
      name: "Final Roughness"
    }),
    metalness: Float($(_templateObject4 || (_templateObject4 = _taggedTemplateLiteral(["patched_Metalness"]))), {
      name: "Final Metalness"
    })
  };
};

export { Blend as B, Layer as L, _taggedTemplateLiteral as _, _toConsumableArray as a, compileModules as c, initialModuleState as i, pipeModules as p };
